import pytest


collect_ignore = ["tests/test_download_builtins.py"]


def pytest_addoption(parser):
    parser.addoption("--no-downloads", action="store_true",
                     default=False,
                     help="skip tests that might download data.")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--no-downloads"):
        skip_test = pytest.mark.skip(
            reason="--no-downloads set"
        )
        for item in items:
            if "might_download" in item.keywords:
                item.add_marker(skip_test)



import pytest

import sncosmo

"""Test all of the builtins code.

This file is designed to test that all of the code that loads builtins works,
and only tests a single example for each kind of builtin. A separate file
(test_download_builtins.py) actually makes sure that all of the builtins can be
downloaded and loaded, but that is slow and requires a lot of downloading so it
isn't included as part of the standard test suite. If you add new builtins
they should be picked up automatically by that file.

You should only add a new test to this file if you created a new loader
function.
"""


@pytest.mark.might_download
def test_builtins_bessell():
    sncosmo.get_bandpass('bessellb')


@pytest.mark.might_download
def test_builtins_remote_aa():
    sncosmo.get_bandpass('standard::u')


@pytest.mark.might_download
def test_builtins_remote_nm():
    sncosmo.get_bandpass('kepler')


@pytest.mark.might_download
def test_builtins_remote_um():
    sncosmo.get_bandpass('f070w')


@pytest.mark.might_download
def test_builtins_remote_wfc3():
    sncosmo.get_bandpass('f098m')


@pytest.mark.might_download
def test_builtins_tophat_um():
    sncosmo.get_bandpass('f1065c')


@pytest.mark.might_download
def test_builtins_megacampsf():
    sncosmo.get_bandpass('megacampsf::u', 0.)


@pytest.mark.might_download
def test_builtins_timeseries_ascii():
    sncosmo.get_source('nugent-sn1a')


@pytest.mark.might_download
def test_builtins_timeseries_fits():
    sncosmo.get_source('hsiao')


@pytest.mark.might_download
def test_builtins_timeseries_fits_local():
    sncosmo.get_source('hsiao-subsampled')


@pytest.mark.might_download
def test_builtins_salt2model():
    sncosmo.get_source('salt2')


@pytest.mark.might_download
def test_builtins_salt3model():
    sncosmo.get_source('salt3')


@pytest.mark.might_download
def test_builtins_2011fe():
    sncosmo.get_source('snf-2011fe')


@pytest.mark.might_download
def test_builtins_mlcs2k2():
    sncosmo.get_source('mlcs2k2')


@pytest.mark.might_download
def test_builtins_snemo():
    sncosmo.get_source('snemo2')


@pytest.mark.might_download
def test_builtins_sugar():
    sncosmo.get_source('sugar')


@pytest.mark.might_download
def test_builtins_magsys_ab():
    sncosmo.get_magsystem('ab')


@pytest.mark.might_download
def test_builtins_magsys_fits():
    sncosmo.get_magsystem('vega')


@pytest.mark.might_download
def test_builtins_magsys_csp():
    sncosmo.get_magsystem('csp')


@pytest.mark.might_download
def test_builtins_magsys_ab_b12():
    sncosmo.get_magsystem('ab-b12')


@pytest.mark.might_download
def test_builtins_magsys_jla():
    sncosmo.get_magsystem('jla1')



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import astropy.units as u
import numpy as np
from scipy.interpolate import splev, splrep

from .bandpasses import get_bandpass
from .constants import HC_ERG_AA, SPECTRUM_BANDFLUX_SPACING, FLAMBDA_UNIT
from .utils import integration_grid

__all__ = ['SpectrumModel']


class SpectrumModel(object):
    """A model spectrum, representing wavelength and spectral density values.

    Parameters
    ----------
    wave : list_like
        Wavelength values.
    flux : list_like
        Spectral flux density values.
    wave_unit : `~astropy.units.Unit`
        Unit on wavelength.
    unit : `~astropy.units.BaseUnit`
        For now, only units with flux density in energy (not photon counts).
    """

    def __init__(self, wave, flux, wave_unit=u.AA,
                 unit=(u.erg / u.s / u.cm**2 / u.AA)):
        self.wave = np.asarray(wave, dtype=np.float64)
        self.flux = np.asarray(flux, dtype=np.float64)
        if self.wave.shape != self.flux.shape:
            raise ValueError('shape of wavelength and flux must match')
        if self.wave.ndim != 1:
            raise ValueError('only 1-d arrays supported')

        # internally, wavelength is in Angstroms:
        if wave_unit != u.AA:
            self.wave = wave_unit.to(u.AA, self.wave, u.spectral())
        self._wave_unit = u.AA

        # internally, flux is in F_lambda:
        if unit != FLAMBDA_UNIT:
            self.flux = unit.to(FLAMBDA_UNIT, self.flux,
                                u.spectral_density(u.AA, self.wave))
        self._unit = FLAMBDA_UNIT

        # Set up interpolation.
        # This appears to be the fastest-evaluating interpolant in
        # scipy.interpolate.
        self._tck = splrep(self.wave, self.flux, k=1)

    def bandflux(self, band):
        """Perform synthentic photometry in a given bandpass.

        The bandpass transmission is interpolated onto the wavelength grid
        of the spectrum. The result is a weighted sum of the spectral flux
        density values (weighted by transmission values).

        Parameters
        ----------
        band : Bandpass or str
            Bandpass object or name of registered bandpass.

        Returns
        -------
        float
            Total flux in ph/s/cm^2.
        """
        band = get_bandpass(band)

        # Check that bandpass wavelength range is fully contained in spectrum
        # wavelength range.
        if (band.minwave() < self.wave[0] or band.maxwave() > self.wave[-1]):
            raise ValueError('bandpass {0!r:s} [{1:.6g}, .., {2:.6g}] '
                             'outside spectral range [{3:.6g}, .., {4:.6g}]'
                             .format(band.name, band.minwave(), band.maxwave(),
                                     self.wave[0], self.wave[-1]))

        # Set up wavelength grid. Spacing (dwave) evenly divides the bandpass,
        # closest to 5 angstroms without going over.
        wave, dwave = integration_grid(band.minwave(), band.maxwave(),
                                       SPECTRUM_BANDFLUX_SPACING)
        trans = band(wave)
        f = splev(wave, self._tck, ext=1)

        return np.sum(wave * trans * f) * dwave / HC_ERG_AA



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test registry functions."""

import numpy as np
import pytest

import sncosmo


def basic_loader(arg, name=None, version=None):
    return arg


def basic_loader_no_version(arg, name=None):
    return arg


@pytest.fixture
def registry():
    # Create a test registry with a few test entries.
    registry = sncosmo._registry.Registry()

    registry.register_loader(
        'test_loader',
        basic_loader,
        args=('test_1',),
        version=1,
    )
    registry.register_loader(
        'test_loader',
        basic_loader,
        args=('test_2',),
        version=2,
        meta={'metaval': 'metatest_2'}
    )
    registry.register('instance_value', 'test_instance')

    return registry


def test_register_bandpass():
    disp = np.array([4000., 4200., 4400., 4600., 4800., 5000.])
    trans = np.array([0., 1., 1., 1., 1., 0.])

    # create a band, register it, make sure we can get it back.
    band = sncosmo.Bandpass(disp, trans, name='tophatg')
    sncosmo.register(band)
    assert sncosmo.get_bandpass('tophatg') is band


def test_register_magsystem():
    magsys = sncosmo.get_magsystem('ab')
    sncosmo.register(magsys, name='test_magsys')
    assert sncosmo.get_magsystem('test_magsys') is magsys


def test_register_source():
    phase = np.linspace(0., 100., 10)
    wave = np.linspace(800., 20000., 100)
    flux = np.ones((len(phase), len(wave)), dtype=float)
    source = sncosmo.TimeSeriesSource(phase, wave, flux, name='test_source')
    sncosmo.register(source)

    # Retrieving a source makes a copy, so just check that the names are the
    # same.
    assert sncosmo.get_source('test_source').name == 'test_source'


def test_register_loader():
    # Register a loader that returns a MagSystem.
    magsys = sncosmo.get_magsystem('ab')
    sncosmo.register_loader(
        sncosmo.magsystems.MagSystem,
        'test_magsystem',
        basic_loader,
        args=(magsys,)
    )

    # make sure that we can get the magnitude system back.
    assert sncosmo.get_magsystem('test_magsystem') is magsys


def test_register_invalid():
    with pytest.raises(ValueError):
        sncosmo.register("invalid")


def test_retrieve():
    sncosmo.registry.retrieve(sncosmo.magsystems.MagSystem, 'ab')


def test_registry_loader_exists(registry):
    with pytest.raises(Exception):
        registry.register_loader(
            'test_loader',
            basic_loader,
            args=('new_test'),
            version=1,
        )


def test_registry_instance_exists(registry):
    with pytest.raises(Exception):
        registry.register('new_instance_value', 'test_instance')


def test_registry_no_name(registry):
    with pytest.raises(ValueError):
        registry.register('asdf')


def test_registry_bad_name(registry):
    class BadClass():
        name = 1
    with pytest.raises(ValueError):
        registry.register(BadClass())


def test_registry_cases(registry):
    # Should work regardless of case.
    for name in ['test_loader', 'TEST_LOADER', 'TesT_LoADeR']:
        registry.retrieve(name)


def test_registry_loader_alias(registry):
    registry.alias('alias_loader', 'test_loader', new_version=1,
                   existing_version=2)
    assert registry.retrieve('alias_loader') == 'test_2'


def test_registry_instance_alias(registry):
    registry.alias('alias_instance', 'test_instance')
    assert registry.retrieve('alias_instance') == 'instance_value'


def test_registry_bad_alias(registry):
    with pytest.raises(Exception):
        registry.alias('alias', 'invalid_key')


def test_registry_retrieve(registry):
    assert registry.retrieve('test_instance') == 'instance_value'


def test_registry_no_version(registry):
    registry.register_loader('noversion_loader', basic_loader_no_version,
                             args=('noversion',))
    assert registry.retrieve('noversion_loader') == 'noversion'


def test_registry_version(registry):
    assert registry.retrieve('test_loader', version=1) == 'test_1'
    assert registry.retrieve('test_loader', version=2) == 'test_2'


def test_registry_default_version(registry):
    assert registry.retrieve('test_loader') == 'test_2'


def test_registry_missing(registry):
    with pytest.raises(Exception):
        assert registry.retrieve('missing')


def test_registry_missing_version(registry):
    with pytest.raises(Exception):
        assert registry.retrieve('test_loader', version=3)


def test_registry_loaders_metadata(registry):
    meta = registry.get_loaders_metadata()
    assert len(meta) == 2



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from collections import OrderedDict


class Registry(object):
    """Connect strings to instances and loaders."""

    def __init__(self):
        self._loaders = OrderedDict()
        self._instances = OrderedDict()
        self._primary_loaders = []  # keys of _loaders not including aliases

    def register_loader(self, name, func, args=None, version=None, meta=None,
                        force=False):
        """Register a data reading function.

        Parameters
        ----------
        name : str
            The data identifier.
        func : callable
            The function to read in the data. Must accept a name and version
            keyword argument.
        args : list, optional
            Arguments to pass to the function. Default is an empty list.
        version : str, optional
            Sub-version of name, if desired. Use formats such as ``'1'``,
            ``'1.0'``, ``'1.0.0'``, etc. Default is `None`.
        force : bool, optional
            Whether to override any existing function if already present.
        meta : dict, optional
            Metadata describing this loader. Default is an empty dictionary.
        """

        if args is None:
            args = []
        if meta is None:
            meta = {}

        name = name.lower()  # names are stored in all-lowercase.
        key = (name, version)

        # check if key already exists in the registry
        if key in self._loaders and not force:
            versionstr = \
                "" if version is None else " (version={!r})".format(version)
            raise Exception("Loader named {!r}{:s} is already "
                            "defined. Use force=True to override."
                            .format(name, versionstr))

        self._loaders[key] = func, args, meta
        self._primary_loaders.append(key)

    def register(self, instance, name=None, force=False):
        """Register a class instance.

        Parameters
        ----------
        instance : object
            The object to be registered.
        name : str, optional
            Identifier. If `None`, the name is taken from the `name` attribute
            of the instance, if it exists and is a string.
        force : bool, optional
            Whether to override any existing instance of the same name.
            Note: this may not play well with versioned instances.
        """

        if name is None:
            try:
                name = instance.name
            except AttributeError:
                raise ValueError("name not given and instance has no 'name' "
                                 "attribute")
            if not isinstance(name, str):
                raise ValueError("name attribute of {0!r:s} is not a string.")

        name = name.lower()
        key = (name, None)

        # check if key is already in instances or loaders:
        if (key in self._instances or key in self._loaders) and not force:
            raise Exception("{0:s} already in registry. Use force=True"
                            " to override.".format(name))

        self._instances[key] = instance

    def alias(self, new_name, existing_name, new_version=None,
              existing_version=None):
        """Alias a new name to an existing name."""

        found = False

        new_key = (new_name, new_version)
        existing_key = (existing_name, existing_version)

        if existing_key in self._loaders:
            found = True
            self._loaders[new_key] = self._loaders[existing_key]

        if existing_key in self._instances:
            found = True
            self._instances[new_key] = self._instances[existing_key]

        if not found:
            raise Exception("{0!r} not found in registry"
                            .format(existing_name))

    def retrieve(self, name, version=None):
        """Retrieve an instance from a registered identifier.

        Parameters
        ----------
        name : str
            Identifier of the specific instance. `name` is case-independent
            (internally names are stored as lowercase).
        version : str
            Sub-identifier. If `None`, default to highest or only version.

        Returns
        -------
        instance

        Notes
        -----
        **Precedence** The following are tried in this order:

        1. If ``name`` is already loaded in the registry, that instance is
           returned.
        2. If there is a loader defined for ``name``, it is used
           to create an instance, save it to the registry and return it.
        3. An Exception is raised listing the available registered names.

        **Versioning** There is support for multiple versions of data for the
        same `name`.

        1. If ``version`` is specified, the registry and its loaders are
           searched for ``(name, version)``.
        2. If ``version`` is not specified but there are registered loaders for
           ``(name, version)``, the latest version is used, and both
           ``(name, None)`` and ``(name, version)`` are saved to the registry.
           "Latest" is defined by string comparision.

        """

        name = name.lower()
        key = (name, version)

        # Try to retrieve from instances
        try:
            return self._instances[key]
        except KeyError:
            pass

        # Try to retrieve from the loaders.
        if key in self._loaders:
            func, args, meta = self._loaders[key]
            if version is None:
                self._instances[key] = func(*args, name=name)
            else:
                self._instances[key] = func(*args, name=name, version=version)
            return self._instances[key]

        # If we got this far and the version is not specified,
        # find the latest version and try to load it.
        if version is None:
            latest_version = None
            for regkey in self._loaders.keys():
                if (name == regkey[0] and (latest_version is None or
                                           regkey[1] > latest_version)):
                    latest_version = regkey[1]
            if latest_version is not None:
                regkey = (name, latest_version)
                func, args, meta = self._loaders[regkey]
                self._instances[regkey] = func(*args, name=name,
                                               version=latest_version)
                self._instances[key] = self._instances[regkey]
                return self._instances[key]

        # At this point we will raise an exception and all the following
        # is to make it more informative.

        # all names regardless of version:
        registered_names = set(k[0] for k in self._loaders)
        registered_names.update(set(k[0] for k in self._instances))

        # If we don't have the name regardless of version:
        if version is None or name not in registered_names:
            raise Exception(
                "{0!r} not in registry. Registered names: '{1}'"
                .format(name, "', '".join(registered_names)))

        # If version was specified but we don't have that specific version:
        registered_versions = [regkey[1] for regkey in self._loaders
                               if name == regkey[0]]
        raise Exception(
            "No {0!r} with version='{1:s}' in registry. Registered"
            " versions: '{2:s}'".format(name, version,
                                        "', '".join(registered_versions)))

    def get_loaders_metadata(self):
        """Return the metadata of all registered loaders.

        Returns
        -------
        loadermeta : list of dict
            Each item in the list is a dictionary containing a 'name'
            keyword, a 'version' keyword (if applicable), and the metadata
            keywords for the given loader.
        """

        result = []
        for key in self._primary_loaders:
            loader = self._loaders[key]
            name, version = key
            m = {'name': name}
            if version is not None:
                m['version'] = version
            m.update(loader[2])
            result.append(m)

        return result
